# Copyright 2010-2018 Lee Harr
#
# This file is part of pynguin.
# https://bitbucket.org/leeharr/pynguin/
#
# Pynguin is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pynguin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pynguin.  If not, see <http://www.gnu.org/licenses/>.


import os
import sys
import shutil
import time
from math import pi
import glob
import zipfile
import logging
logger = logging.getLogger('PynguinLogger')

from .bidict import bidict


from PyQt5 import QtGui, QtCore, QtWidgets, uic
from PyQt5.Qt import QHBoxLayout

from .pynguin import Pynguin, pynguin_functions, interpreter_protect, PynThread
from .mode import ModeLogo, ModeTurtle
from . import util
from .util import sign, get_datadir, get_docdir
from .ed4 import Documents, Editor
from .interpreter import Interpreter, CmdThread, Console, WatcherThread
from .about import AboutDialog
from . import conf

datadir = get_datadir()
uidir = os.path.join(datadir, 'ui')
docdir = get_docdir()


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, app):
        self.app = app
        self.paused = False
        self.svgrenderer = util.SvgRenderer(self.app)
        self.rend = self.svgrenderer.getrend()

        self._fdir_default = os.path.expanduser('~')
        self._fdir = self._fdir_default
        self._filepath = None
        self._modified = False

        self._mainthread = QtCore.QThread.currentThread()
        #logger.info('THREADS')
        #logger.info(self._mainthread)
        #import threading
        #logger.info(threading.current_thread())

        uifile = 'pynguin.ui'
        uipath = os.path.join(uidir, uifile)
        MWClass, _ = uic.loadUiType(uipath)

        QtWidgets.QMainWindow.__init__(self)
        self.ui = MWClass()
        self.ui.setupUi(self)

        self.initialize_settings()
        self.setup_avatar_choices()

        self.scene = Scene()
        view = self.ui.view
        view.setScene(self.scene)
        self.scene.view = view
        view.show()
        view.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        view.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        view.mousePressEvent = self.onclick
        view.wheelEvent = self.mousewheelscroll
        view.mouseMoveEvent = self.mousemove
        view.mouseReleaseEvent = self.mouserelease

        self.setup_speed_choices()
        self.speedgroup = QtWidgets.QActionGroup(self)
        self.speedgroup.addAction(self.ui.actionSlow)
        self.speedgroup.addAction(self.ui.actionMedium)
        self.speedgroup.addAction(self.ui.actionFast)
        self.speedgroup.addAction(self.ui.actionInstant)
        self.speedgroup.setExclusive(True)
        self.speedgroup.triggered.connect(self.speedMenuEvent)

        hbox = QHBoxLayout(self.ui.edframe)
        hbox.setSpacing(0)
        hbox.setContentsMargins(0, 0, 0, 0)
        self.editor_box = hbox
        self.editor = Documents(self, Editor)
        self.editor_box.addLayout(self.editor.box)

        self.interpretereditor = Interpreter(self)
        hbox = QHBoxLayout(self.ui.interpreterframe)
        hbox.setSpacing(0)
        hbox.setContentsMargins(0, 0, 0, 0)
        hbox.addWidget(self.interpretereditor)
        ilocals = {}
        self.interpreter_locals = ilocals
        self.interpreter = Console(self.interpreter_locals, self.interpretereditor)
        self.interpretereditor.interpreter = self.interpreter
        self.interpretereditor.setFocus()

        self._setup_pendown_choices()
        self.pengroup = QtWidgets.QActionGroup(self)
        self.pengroup.addAction(self.ui.actionPenUp)
        self.pengroup.addAction(self.ui.actionPenDown)
        self.pengroup.setExclusive(True)
        self.pengroup.triggered.connect(self.setPen)

        Pynguin.mw = self
        Pynguin.rend = self.rend
        self.pynguins = [] # User-code-facing list of pynguins
        self._pynguins = [] # Actual list of existing pynguins
        self._defunct_pynguins = []
        self.pynguin = None
        try:
            self.pynguin = self.new_pynguin()
        except:
            settings = QtCore.QSettings()
            settings.clear()
            self.pynguin = self.new_pynguin()

        self._scale = 1
        trans = QtGui.QTransform()
        trans.scale(self._scale, self._scale)
        self.scene.view.setTransform(trans)
        view.centerOn(self.pynguin.gitem)

        self.interpretereditor.pynguin = self.pynguin
        self.setup_interpreter_locals()

        self.ui.rsplitter.setSizes([390, 110])
        self.ui.wsplitter.setSizes([550, 350])
        self.ui.wsplitter.splitterMoved.connect(self.recenter)

        self.viewgroup = QtWidgets.QActionGroup(self)
        self.viewgroup.addAction(self.ui.actionPynguin)
        self.viewgroup.addAction(self.ui.actionArrow)
        self.viewgroup.addAction(self.ui.actionRobot)
        self.viewgroup.addAction(self.ui.actionTurtle)
        self.viewgroup.addAction(self.ui.actionHidden)
        self.viewgroup.setExclusive(True)
        self.viewgroup.triggered.connect(self.setImageEvent)

        self.modegroup = QtWidgets.QActionGroup(self)
        self.modegroup.addAction(self.ui.actionModePynguin)
        self.modegroup.addAction(self.ui.actionModeLogo)
        self.modegroup.addAction(self.ui.actionModeTurtle)
        self.modegroup.setExclusive(True)

        self._setup_pendown_choices()
        self.pengroup = QtWidgets.QActionGroup(self)
        self.pengroup.addAction(self.ui.actionPenUp)
        self.pengroup.addAction(self.ui.actionPenDown)
        self.pengroup.setExclusive(True)
        self.pengroup.triggered.connect(self.setPen)

        self._setup_fill_choices()
        self.fillgroup = QtWidgets.QActionGroup(self)
        self.fillgroup.addAction(self.ui.actionFill)
        self.fillgroup.addAction(self.ui.actionNofill)
        self.fillgroup.setExclusive(True)
        self.fillgroup.triggered.connect(self.setFill)

        self._bgcolor = None

        self.setup_settings()
        self.setup_recent()

        self.setup_examples()

        QtCore.QTimer.singleShot(60000, self.autosave)
        self._centerview()
        self._leftdragstart = None

        self.watcher = QtCore.QFileSystemWatcher(self)
        self.watcher.fileChanged.connect(self._filechanged)
        self._watchdocs = {}
        self._writing_external = None

        self._movetimer = self.startTimer(0)


    def setup_interpreter_locals(self, pyn=None):
        '''insert some values in to the interpreter.

        By default, inserts the "main" pynguin and its methods, but
            can pass in another pynguin and use it and its methods
            instead. (Useful when switching over from one pynguin
            to another as the main pynguin, like in reap() or promote()
        '''

        if pyn is None:
            pyn = self.pynguin

        ilocals = self.interpreter_locals
        ilocals.update(PI=pi,
                        Pynguin=Pynguin,
                        ModeLogo=ModeLogo,
                        ModeTurtle=ModeTurtle,
                        pynguin=pyn,
                        p=pyn,
                        pynguins=self.pynguins,
                        history=self.history,
                        log=logger.info,
                        util=util,
                        Thread=PynThread)

        for fname in pynguin_functions:
            function = getattr(pyn, fname)
            ilocals[fname] = function

    def mousewheelscroll(self, ev):
        # only supports old-style mouse wheel
        # ... could do something different on a "sideways" scroll?
        xydelta = ev.angleDelta()
        ydelta = xydelta.y()
        view = self.scene.view
        view.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.zoom(ydelta)
        view.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorViewCenter)

    def zoom(self, delta):
        view = self.scene.view

        scaleperc = 1 + ((delta / 120.0) * 0.05)
        self._scale *= scaleperc

        trans = QtGui.QTransform()
        trans.scale(self._scale, self._scale)
        view.setTransform(trans)

        self._centerview()

    def zoomin(self):
        self.zoom(120)
    def zoomout(self):
        self.zoom(-120)
    def zoom100(self):
        self._scale = 1
        self.zoom(0)
        pynguin = self.pynguin
        if hasattr(pynguin, '_pyn'):
            gitem = pynguin._pyn.gitem
        else:
            gitem = pynguin.gitem
        self.scene.view.ensureVisible(gitem)
    def zoomfit(self):
        'Zoom to fit the whole drawing'

        fitrect = self.scene.itemsBoundingRect()
        self.scene.view.fitInView(fitrect, QtCore.Qt.KeepAspectRatio)
        tr = self.scene.view.transform()
        self._scale = tr.m11()
        self.zoom(0)

    def mousemove(self, ev):
        QtWidgets.QGraphicsView.mouseMoveEvent(self.scene.view, ev)

        if self._leftdragstart is None:
            ev.accept()
            return

        buttons = ev.buttons()
        if buttons & QtCore.Qt.LeftButton:
            pos = ev.pos()
            dpos = self._leftdragstart - pos
            ctr0 = self._dragstartcenter
            ctr = ctr0 + (dpos / self._scale)

            self._cx = ctr.x()
            self._cy = ctr.y()
            self.recenter()

    def pan(self, dx=0, dy=0):
        cx0 = self._cx
        cy0 = self._cy
        self._cx +=  dx / self._scale
        self._cy -=  dy / self._scale
        self.recenter()
        ctr1 = self._viewcenter()
        cx1 = ctr1.x()
        cy1 = ctr1.y()
        if dx:
            ctr = QtCore.QPointF(cx1, cy0)
        elif dy:
            ctr = QtCore.QPointF(cx0, cy1)
        self._centerview(ctr)

    def panleft(self):
        self.pan(dx=-25)
    def panright(self):
        self.pan(dx=+25)
    def panup(self):
        self.pan(dy=+25)
    def pandown(self):
        self.pan(dy=-25)

    def _viewrect(self):
        view = self.scene.view
        viewportrect = view.viewport().geometry()
        tl = viewportrect.topLeft()
        br = viewportrect.bottomRight()
        tlt = view.mapToScene(tl)
        brt = view.mapToScene(br)
        return QtCore.QRectF(tlt, brt)

    def _viewcenter(self):
        'return the current center of the view'
        ctr = self._viewrect().center()
        return ctr

    def _centerview(self, ctr=None):
        'set _cx and _cy to the given coord or to the current view center'
        if ctr is None:
            ctr = self._viewcenter()
        self._cx = ctr.x()
        self._cy = ctr.y()

    def _sync_track(self, on=True):
        self.ui.actionTrack.setChecked(on)
        settings = QtCore.QSettings()
        settings.setValue('pynguin/track', on)
    def settrack(self):
        track = self.ui.actionTrack.isChecked()
        self.pynguin.track(track)
        if track:
            cmd = 'track()\n'
        else:
            cmd = 'notrack()\n'
        self.interpretereditor.addcmd(cmd)

    def onclick(self, ev):
        QtWidgets.QGraphicsView.mousePressEvent(self.scene.view, ev)

        button = ev.button()
        if not ev.isAccepted() and button==QtCore.Qt.LeftButton:
            self.leftclick(ev)

        elif button==QtCore.Qt.RightButton:
            self.rightclick(ev)

        self.interpretereditor.setFocus()

    def leftclick(self, ev):
        #print(dir(ev))
        if not ev.isAccepted():
            self._leftdragstart = ev.pos()
            self._dragstartcenter = self._viewcenter()
        else:
            self._leftdragstart = None

    def mouserelease(self, ev):
        QtWidgets.QGraphicsView.mouseReleaseEvent(self.scene.view, ev)
        self._leftdragstart = None

    def rightclick(self, ev):
        evpos = ev.pos()
        scpos = self.scene.view.mapToScene(evpos)
        for pyn in self.pynguins:
            if pyn is self.pynguin and 'onclick' in self.interpreter_locals:
                onclick = self.interpreter_locals['onclick']
            else:
                onclick = pyn.onclick
            onclick(scpos.x(), scpos.y())
        ev.ignore()

    def recenter(self):
        center = QtCore.QPointF(self._cx, self._cy)
        self.ui.view.centerOn(center)

    def setup_examples(self):
        filemenu = self.ui.filemenu
        actionsave = self.ui.actionSave
        exmenu = QtWidgets.QMenu('Examples', filemenu)
        exa = filemenu.insertMenu(actionsave, exmenu)
        examplemenu = exa.menu()

        examplesrel = 'examples'
        examplespath = os.path.join(docdir, examplesrel)
        self.examplespath = examplespath
        examplesglob = '%s/*.pyn' % examplespath
        examples = glob.glob(examplesglob)

        # push some of the examples to the top of the list
        ordered = ['welcome', 'somex', 'morex']
        topex = []
        for o in ordered:
            fn = '%s.pyn' % o
            for ex in examples:
                if ex.endswith(fn):
                    topex.append(ex)
        for place, ex in enumerate(topex):
            examples.remove(ex)
            examples.insert(place, ex)

        for ex in examples:
            pth, fn = os.path.split(ex)
            def excb(fp=ex):
                self.open_example(fp)
            action = examplemenu.addAction(fn, excb)

    def open_example(self, fp):
        if not self.maybe_save():
            return
        else:
            self._filepath = None

        self._new()
        self._openfile(fp, False)
        self._fformat = None
        self.interpreter.editor.set_commandsmodify(False)

    def settings(self):
        'open the settings dialog'
        from . import settings
        s = settings.Settings(self)
        r = s.exec_()
        if r:
            settings = QtCore.QSettings()
            ui = s.ui

            savesingle = ui.savesingle.isChecked()
            oldsavesingle = settings.value('file/savesingle', True, bool)
            if savesingle != oldsavesingle:
                if self._filepath is None:
                    # example file
                    pass
                elif savesingle and self._fformat=='.pyn':
                    # new format matches file format
                    pass
                elif not savesingle and self._fformat=='_pynd':
                    # new format matches file format
                    pass
                else:
                    # Need to mark all docs as modified to make sure
                    # they get written out in the new kind of save file
                    mselect = self.ui.mselect
                    for doc in self.editor.documents.values():
                        doc.setModified(True)
                    self.editor._modified = True
                    self._modified = True
                    self.show_modified_status()
                    self._fformat_asked = False
            settings.setValue('file/savesingle', savesingle)

            reloadexternal = ui.reloadexternal.isChecked()
            settings.setValue('file/reloadexternal', reloadexternal)
            autorun = ui.autorun.isChecked()
            settings.setValue('file/autorun', autorun)

            bfp = ui.backupfolderpath.text()
            settings.setValue('file/backupfolderpath', bfp)
            bfn = ui.backupfilename.text()
            settings.setValue('file/backupfilename', bfn)
            brate = ui.backuprate.value()
            settings.setValue('file/backuprate', brate)
            # hold on to the old backupkeep value. If it was zero
            # need to start the autosave timer
            bkeep0 = settings.value('file/backupkeep', 5, int)
            bkeep = ui.backupkeep.value()
            settings.setValue('file/backupkeep', bkeep)
            if not bkeep0 and bkeep:
                QtCore.QTimer.singleShot(brate*60000, self.autosave)

            reset = ui.testrun_reset.isChecked()
            settings.setValue('editor/testrun_reset', reset)
            mainfirst = ui.editor_mainfirst.isChecked()
            settings.setValue('editor/mainfirst', mainfirst)
            rev = ui.testall_reverse.isChecked()
            settings.setValue('editor/testall_reverse', rev)
            autocall = ui.testall_autocall.isChecked()
            settings.setValue('editor/testall_autocall', autocall)

            bgcolor = ui._bgcolor
            settings.setValue('view/bgcolor', bgcolor.name())
            color = ui._color
            settings.setValue('pynguin/color', color.rgba())
            fillcolor = ui._fillcolor
            settings.setValue('pynguin/fillcolor', fillcolor.rgba())

            reset_forces_visible = ui.reset_forces_visible.isChecked()
            settings.setValue('pynguin/reset_forces_visible', reset_forces_visible)
            allow_start_hidden = ui.allow_start_hidden.isChecked()
            settings.setValue('pynguin/allow_start_hidden', allow_start_hidden)

            quiet = ui.quietinterrupt.isChecked()
            settings.setValue('console/quietinterrupt', quiet)
            commandsmodify = ui.commandsmodify.isChecked()
            settings.setValue('console/commandsmodify', commandsmodify)
            self.interpreter.editor.commandsmodify = commandsmodify

    def clear_settings(self):
        '''reset settings to default values.
        '''
        settings = QtCore.QSettings()
        settings.clear()
        self.setup_settings()
        self.pynguin.mode('pynguin')
        self.pynguin.reset(True)

    def initialize_settings(self):
        QtCore.QCoreApplication.setOrganizationName('pynguin.googlecode.com')
        QtCore.QCoreApplication.setOrganizationDomain('pynguin.googlecode.com')
        QtCore.QCoreApplication.setApplicationName('pynguin')

    def setup_settings(self):
        settings = QtCore.QSettings()
        self.settings = settings

        geometry = settings.value('window/geometry', None)
        if geometry is not None:
            self.restoreGeometry(geometry)

        self.pynguin._set_bgcolor_to_default()
        self.pynguin._set_color_to_default()
        self.pynguin._set_fillcolor_to_default()

        fontsize = settings.value('editor/fontsize', 0, int)
        self.editor.setfontsize(fontsize)
        wrap = settings.value('editor/wordwrap', False, bool)
        if wrap:
            self.ui.actionWordwrap.setChecked(True)
            self.wordwrap()
        numbers = settings.value('editor/linenumbers', True, bool)
        if not numbers:
            self.ui.actionShowLineNumbers.setChecked(False)
            self.editor.line_numbers(False)

        default = 'fast'
        speed = settings.value('pynguin/speed', default)
        try:
            self.pynguin.speed(speed)
        except ValueError:
            self.pynguin.speed(default)

        # remember the saved avatar
        imageid = settings.value('pynguin/avatar', 'pynguin')
        allow_start_hidden = settings.value('pynguin/allow_start_hidden', False, bool)
        if not allow_start_hidden and imageid=='hidden':
            imageid = 'pynguin'

        # set up any custom avatars
        n = settings.beginReadArray('pynguin/custom_avatars')
        for i in range(n):
            settings.setArrayIndex(i)
            idp = settings.value('idpath')
            self.set_pynguin_avatar(idp)
        settings.endArray()

        # then set the saved avatar that we remembered earlier
        self.set_pynguin_avatar(imageid)
        self._sync_avatar_menu(imageid)

        track = settings.value('pynguin/track', False, bool)
        if track:
            self.pynguin.track()

        self.interpreter.editor.set_commandsmodify()

    def setup_recent(self):
        settings = self.settings
        settings = QtCore.QSettings()
        recent = []
        for n in range(settings.beginReadArray('recent')):
            settings.setArrayIndex(n)
            fname = settings.value('fname')
            fnamestr = str(fname)
            if os.path.exists(fnamestr):
                recent.append(str(fname))
        settings.endArray()

        filemenu = self.ui.filemenu
        if not hasattr(self.ui, 'recmenu'):
            actionsave = self.ui.actionSave
            rmenu = QtWidgets.QMenu('Recent', filemenu)
            rec = filemenu.insertMenu(actionsave, rmenu)
            recmenu = rec.menu()
            self.ui.recmenu = recmenu
        else:
            recmenu = self.ui.recmenu
            recmenu.clear()

        for fp in recent:
            pth, fn = os.path.split(fp)
            if not fn:
                continue
            def excb(fp=fp):
                self.open_recent(fp)
            recmenu.addAction(fn, excb)

    def open_recent(self, fp):
        if not os.path.exists(fp):
            QtWidgets.QMessageBox.information(self,
                    'File not found',
                    'File not found:\n\n%s' % fp)
            self.rmrecentfile(fp)
            return

        if fp.endswith('.py'):
            pass
        elif not self.maybe_save():
            return
        else:
            self._filepath = None

        fdir, _ = os.path.split(fp)
        oldfilepath = self._filepath
        if fp.endswith('.pyn'):
            self._new()
            self._filepath = fp
            success = self._openfile(fp)
        elif fp.endswith('.py'):
            success = self._openfile(fp)
        elif fp.endswith('_pynd'):
            self._new()
            self._filepath = fp
            success = self._opendir(fp)
        else:
            try:
                success = self._openfile(fp)
            except:
                try:
                    success = self._opendir(fp)
                except:
                    QtWidgets.QMessageBox.information(self,
                            'Open failed',
                            'Unable to open file:\n\n%s' % fp)

        if success:
            self._fdir = fdir
        else:
            self._filepath = oldfilepath

    def new_pynguin(self, mname=None, show_cmd=True):
        settings = QtCore.QSettings()
        if not mname:
            # remember the saved mode
            mname = settings.value('pynguin/mode', 'pynguin')

        from . import mode
        class_name = mode.modes[mname]

        cls = globals()[class_name]
        p = cls()
        p._modename = mname

        ilocals = self.interpreter_locals
        if 'p' not in ilocals:
            ilocals['p'] = p
            self._sync_mode_menu(mname)

        elif show_cmd:
            pn = 2
            while True:
                pns = 'p%s' % pn
                if pns not in ilocals:
                    break
                pn += 1
            ilocals[pns] = p
            cmd = '%s = %s()\n' % (pns, class_name)
            self.interpretereditor.addcmd(cmd)

        imageid = settings.value('pynguin/avatar', 'pynguin')
        self.set_pynguin_avatar(imageid, p)

        return p

    def reset_pynguin(self):
        self.pynguin.reset()
        cmd = 'reset()\n'
        self.interpretereditor.addcmd(cmd)

    def clear_canvas(self):
        self.pynguin.clear()
        cmd = 'clear()\n'
        self.interpretereditor.addcmd(cmd)

    def set_mode_logo(self):
        self.pynguin.mode('logo')

        cmd = "mode('logo')\n"
        self.interpretereditor.addcmd(cmd)

    def set_mode_turtle(self):
        self.pynguin.mode('turtle')

        cmd = "mode('turtle')\n"
        self.interpretereditor.addcmd(cmd)

    def set_mode_pynguin(self):
        self.pynguin.mode('pynguin')

        cmd = "mode('pynguin')\n"
        self.interpretereditor.addcmd(cmd)

    def _sync_mode_menu(self, mname):
        logger.info('MN %s' % mname)
        choices = {
            'pynguin': self.ui.actionModePynguin,
            'logo': self.ui.actionModeLogo,
            'turtle': self.ui.actionModeTurtle}

        action = choices[mname]
        action.setChecked(True)

    def timerEvent(self, ev):
        Pynguin._process_moves()

    def closeEvent(self, ev=None):
        if self.maybe_save():
            self.interpretereditor.cmdthread = None
            settings = QtCore.QSettings()
            settings.setValue('window/geometry', self.saveGeometry())
            ev.accept()
        else:
            ev.ignore()

    def history(self, clear=False):
        if not clear:
            for line in self.interpretereditor.history:
                self.interpretereditor.write('%s\n' % line)
        else:
            self.interpretereditor.history = []

    def new(self):
        if self.maybe_save():
            self._new(newdoc=True)
        else:
            pass

    def _new(self, newdoc=False):
        import threading
        if hasattr(threading, 'threads'):
            for pyn in threading.threads:
                threading.threads[pyn] = 0

        self.pynguin.reset(True)
        del_later = []
        for name in self.interpreter_locals:
            if name not in pynguin_functions and name not in interpreter_protect:
                del_later.append(name)
        for name in del_later:
            del self.interpreter_locals[name]
        self.editor.clear_all()
        self.interpretereditor.clear()
        if newdoc:
            self.newdoc()
        self._modified = False
        windowtitle = 'pynguin [*]'
        self.setWindowTitle(windowtitle)
        self.show_modified_status()
        self._filepath = None

    def check_modified(self):
        if self._modified:
            return True
        elif self.editor._modified:
            return True
        else:
            for tdoc in list(self.editor.documents.values()):
                if tdoc.isModified():
                    return True

        return False

    def show_modified_status(self):
        if self.check_modified():
            self.setWindowModified(True)
        else:
            self.setWindowModified(False)

    def maybe_save(self):
        if self.check_modified():
            ret = QtWidgets.QMessageBox.warning(self,
                    self.tr('Save Changes?'),
                    self.tr("The document has been modified.\n"
                            "Do you want to save your changes?"),
                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No | QtWidgets.QMessageBox.Cancel,
                    QtWidgets.QMessageBox.Cancel)
            if ret == QtWidgets.QMessageBox.Yes:
                return self.save()
            elif ret == QtWidgets.QMessageBox.Cancel:
                return False
        return True

    def autosave(self):
        '''automatically save a copy of the current project.
        Also moves older backups up a number and deletes the
            oldest backup.
        '''
        settings = QtCore.QSettings()
        backupkeep = settings.value('file/backupkeep', 5, int)
        if not backupkeep:
            return

        bfp = settings.value('file/backupfolderpath', '')
        bfp = bfp or self._fdir_default
        bfn = settings.value('file/backupfilename', conf.BACKUP_FILENAME)
        fp = os.path.join(bfp, bfn % '')
        if self.writeable(fp, savesingle=True):
            self._writefile01(fp, backup=True)
        else:
            QtWidgets.QMessageBox.warning(self,
                    'Autosave failed',
                    '''Cannot complete autosave.
Autosave disabled.
Check configuration!''')
            settings.setValue('file/backupkeep', 0)
            return

        if backupkeep > 1:
            for fn in range(backupkeep, 1, -1):
                fpsrc = os.path.join(bfp, bfn % (fn-1))
                fpdst = os.path.join(bfp, bfn % fn)
                if os.path.exists(fpsrc):
                    shutil.move(fpsrc, fpdst)
        else:
            fpsrc = os.path.join(bfp, bfn % 1)

        shutil.move(fp, fpsrc)

        brate = settings.value('file/backuprate', 3, int)
        QtCore.QTimer.singleShot(brate*60000, self.autosave)

    def _correct_filename(self, fp=None):
        'make sure file name ends with .pyn'
        if fp is None:
            fp = self._filepath
        r, ext = os.path.splitext(fp)
        if ext != '.pyn':
            ext = '.pyn'
        return r + ext

    def _writefile(self, fp):
        'Write file list files and history as a zip file called .pyn'
        z = zipfile.ZipFile(fp, 'w')

        mselect = self.ui.mselect
        for n in range(mselect.count()):
            docid = str(mselect.itemData(n))
            doc = self.editor.documents[docid]
            modified = doc.isModified()
            cline, ccol = doc.getCursorPosition()
            arcname = '##%05d##__%s' % (n, docid)
            code = doc.cleancode()
            doc.beginUndoAction()
            doc.selectAll()
            doc.removeSelectedText()
            doc.insert(code)
            doc.setCursorPosition(cline, ccol)
            doc.setModified(modified)
            doc.endUndoAction()
            z.writestr(arcname, code.encode('utf-8'))

        historyname = '@@history@@'
        history = '\n'.join(self.interpretereditor.history)
        z.writestr(historyname, history.encode('utf-8'))

        z.close()

    def _writefile01(self, fp, backup=False):
        '''Write file list files and history as a zip file called *.pyn

        Version 01

        Puts all files in to a folder to make it match the writedir
        method better.

        Also used by _writedir() but all internal files that should be
        written to the related directory will be prefaced with '_@@'

        '''
        VERSION = b'pyn01'

        _, fn = os.path.split(fp)
        fbase, _ = os.path.splitext(fn)
        if not fbase.endswith('_pynd'):
            fbase = fbase + '_pynd'

        z = zipfile.ZipFile(fp, 'w')

        manifest = []

        mselect = self.ui.mselect
        for n in range(mselect.count()):
            docid = str(mselect.itemData(n))
            doc = self.editor.documents[docid]
            modified = doc.isModified()
            cline, ccol = doc.getCursorPosition()
            code = doc.cleancode()
            doc.beginUndoAction()
            doc.selectAll()
            doc.removeSelectedText()
            doc.insert(code)
            doc.setCursorPosition(cline, ccol)
            doc.setModified(modified)
            doc.endUndoAction()

            logger.info('writing %s' % n)
            if not backup and doc._filepath is not None:
                efp = doc._filepath
                manifest.append(efp)

                logger.info('separate: %s' % efp)

                if efp.startswith('_@@'):
                    dirname, _ = os.path.split(fp)
                    if dirname:
                        efp = efp.replace('_@@', dirname)
                    else:
                        efp = efp[4:]
                    logger.info('internal: %s' % efp)
                else:
                    logger.info('external: %s' % efp)
                self._remwatcher(efp)
                f = open(efp, 'w')
                f.write(code)
                f.close()
                self._addwatcher(efp, doc)

                doc.setModified(False)

            else:
                arcname = '%05d.py' % n

                zipi = zipfile.ZipInfo()
                zipi.filename = '/'.join([fbase, arcname])
                manifest.append(zipi.filename)
                zipi.compress_type = zipfile.ZIP_DEFLATED
                zipi.date_time = time.localtime()
                zipi.external_attr = 0o644 << 16
                zipi.comment = VERSION
                z.writestr(zipi, code.encode('utf-8'))

        historyname = '@@history@@'
        history = '\n'.join(self.interpretereditor.history)

        zipi = zipfile.ZipInfo()
        zipi.filename = os.path.join(fbase, historyname)
        zipi.compress_type = zipfile.ZIP_DEFLATED
        zipi.date_time = time.localtime()
        zipi.external_attr = 0o644 << 16
        zipi.comment = VERSION
        z.writestr(zipi, history) #.encode('utf-8'))

        manifestname = '@@manifest@@'
        zipi = zipfile.ZipInfo()
        zipi.filename = os.path.join(fbase, manifestname)
        zipi.compress_type = zipfile.ZIP_DEFLATED
        zipi.date_time = time.localtime()
        zipi.external_attr = 0o644 << 16
        zipi.comment = VERSION
        manifeststr = '\n'.join(manifest)
        z.writestr(zipi, manifeststr.encode('utf-8'))

        z.close()

        self.show_modified_status()

    def _related_dir(self, fp=None):
        '''return the directory name related to path fp
        This is the directory that will be used for storing
            the files when not in savesingle mode.
        '''
        if fp is None:
            fp = self._filepath

        dirname, fname = os.path.split(fp)
        fbase, fext = os.path.splitext(fname)
        fpd = os.path.join(dirname, fbase+'_pynd')

        return fpd

    def _writedir(self, fp):
        'Write file list files in to a directory'

        logger.info('_writedir %s' % fp)
        fpd = self._related_dir(fp)
        logger.info('RELATED: %s' % fpd)
        _, dirname = os.path.split(fpd)

        if not os.path.exists(fpd):
            os.mkdir(fpd)

        mselect = self.ui.mselect
        count = mselect.count()
        for n in range(count):
            docid = str(mselect.itemData(n))
            doc = self.editor.documents[docid]
            if doc._filepath is None:
                arcname = '%05d.py' % n
                doc._filepath = os.path.join('_@@', dirname, arcname)
                doc.setModified(True)
                logger.info('dfp %s' % doc._filepath)
            else:
                logger.info('DFP %s' % doc._filepath)

        self._writefile01(fp)

        for n in range(count):
            docid = str(mselect.itemData(n))
            doc = self.editor.documents[docid]
            if doc._filepath.startswith('_@@'):
                doc._filepath = None

    def _savestate(self):
        '''write out the files in the editor window, and keep the list
            of history entries.

        if configured to use a single file:
            All of this is packed up in to a zip
            file and given a .pyn filename ending.

        if configured for directory of files:
            The files will be saved in a separate directory.
            The directory will end in _pynd
            A .pyn file will also be saved indexing the contents and
                allowing the files to be easily re-loaded.

        '''

        settings = QtCore.QSettings()
        savesingle = settings.value('file/savesingle', True, bool)

        newfmt = None
        newfmtpyn = False
        newfmtpynd = False
        asked = False

        if hasattr(self, '_fformat_asked') and self._fformat_asked:
            # user has already been asked about save format
            fmt = self._fformat
            asked = True
        elif hasattr(self, '_fformat') and self._fformat:
            # document has a format set
            fmt = self._fformat
        elif savesingle:
            # from settings
            fmt = '.pyn'
        else:
            # from settings
            fmt = '_pynd'

        if asked and fmt=='.pyn':
            newfmt = fmt
            newfmtpyn = True

        elif asked and fmt=='_pynd':
            newfmt = fmt
            newfmtpynd = True

        elif savesingle and fmt!='.pyn':
            logger.debug('mismatch0')
            from . import fformat
            dlg = fformat.FFormat(self)
            dlg.ui.message.setText("Document was previous saved in .pyn + _pynd directory format.\n"
                        "Do you want to save in .pyn archive format?")
            dlg.ui.pyn.setChecked(True)
            ret = dlg.exec_()

            if not ret:
                return False

            newfmtpyn = dlg.ui.pyn.isChecked()
            newfmtpynd = dlg.ui.pynd.isChecked()
            if newfmtpyn:
                newfmt = '.pyn'
            elif newfmtpynd:
                newfmt = '_pynd'
            else:
                raise ValueError
            self._fformat_asked = True

        elif not savesingle and fmt!='_pynd':
            logger.debug('mismatch1')
            from . import fformat
            dlg = fformat.FFormat(self)
            dlg.ui.message.setText("Document was previously saved in .pyn archive format.\n"
                        "Do you want to save in .pyn + _pynd directory format?")
            dlg.ui.pynd.setChecked(True)
            ret = dlg.exec_()

            if not ret:
                return False

            newfmtpyn = dlg.ui.pyn.isChecked()
            newfmtpynd = dlg.ui.pynd.isChecked()
            if newfmtpyn:
                newfmt = '.pyn'
            elif newfmtpynd:
                newfmt = '_pynd'
            else:
                raise ValueError
            self._fformat_asked = True

        if self.writeable(self._filepath):
            self._fformat = newfmt or fmt

            if newfmtpyn or (savesingle and not newfmtpynd):
                self._writefile01(self._filepath)
            else:
                self._writedir(self._filepath)

        else:
            QtWidgets.QMessageBox.warning(self,
                    'Save failed',
                    '''Cannot write to selected file.''')
            self._filepath = None
            return False

        self._modified = False
        for tdoc in list(self.editor.documents.values()):
            tdoc.setModified(False)
        self.editor._modified = False
        self.show_modified_status()

        self.addrecentfile(self._filepath)

        if savesingle:
            self._verify_saved(self._filepath)

        return True


    def writerecentfiles(self, recent):
        recent = recent[:6]

        settings = QtCore.QSettings()
        settings.beginWriteArray('recent')
        for n in range(len(recent)):
            settings.setArrayIndex(n)
            fpath = recent[n]
            settings.setValue('fname', fpath)
        settings.endArray()
        self.setup_recent()

    def addrecentfile(self, fp):
        settings = QtCore.QSettings()
        recent = [fp]
        for n in range(settings.beginReadArray('recent')):
            settings.setArrayIndex(n)
            fname = str(settings.value('fname'))
            if fname and fname not in recent:
                recent.append(fname)
        settings.endArray()
        self.writerecentfiles(recent)

    def rmrecentfile(self, fp):
        settings = QtCore.QSettings()
        recent = []
        for n in range(settings.beginReadArray('recent')):
            settings.setArrayIndex(n)
            fname = str(settings.value('fname'))
            if fname and fname not in recent and fname != fp:
                recent.append(fname)
        settings.endArray()
        self.writerecentfiles(recent)

    def save(self):
        '''call _savestate with current file name, or get a file name from
            the user and then call _savestate
        '''

        if self._filepath is None:
            return self.saveas()
        else:
            return self._savestate()

    def _make_related_dir(self, fp):
        fpd = self._related_dir(fp)
        if not os.path.exists(fpd):
            os.mkdir(fpd)
        elif os.path.exists(fpd) and not self.overwrite(fpd):
            return False
        elif os.listdir(fpd):
            import shutil
            shutil.rmtree(fpd)
            os.mkdir(fpd)
        return True

    def saveas(self):
        if self._fdir is None:
            fdir = os.path.abspath(os.path.curdir)
        else:
            fdir = self._fdir

        fp, _ = QtWidgets.QFileDialog.getSaveFileName(self,
                        'Save As', fdir,
                        'Text files (*.pyn)')

        fp = str(fp)
        if fp:
            settings = QtCore.QSettings()
            savesingle = settings.value('file/savesingle', True, bool)

            fp = self._correct_filename(fp)
            dirname, fname = os.path.split(fp)
            if not savesingle:
                if not self._make_related_dir(fp):
                    return False

        else:
            return False

        oldfilepath = self._filepath
        self._filepath = fp
        retval = self._savestate()

        if retval:
            fdir, fname = os.path.split(self._filepath)
            windowtitle = '%s [*] - Pynguin' % fname
            self.setWindowTitle(windowtitle)
            self.show_modified_status()

            self._fdir = dirname
        else:
            self._filepath = oldfilepath

        return retval

    def overwrite(self, fname):
        '''ask the user if they want to overwrite an existing file
            or directory.

        '''
        ret = QtWidgets.QMessageBox.warning(self,
                self.tr('Overwrite?'),
                self.tr("The Directory already exists:\n%s\n"
                        "Do you want to delete it and all contents?" % fname),
                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No | QtWidgets.QMessageBox.Cancel,
                    QtWidgets.QMessageBox.Cancel)
        if ret == QtWidgets.QMessageBox.Yes:
            return True
        else:
            return False

    def writeable(self, fp, savesingle=None):
        'try to write to the given path. return True if sucessful.'
        try:
            fd = open(fp, 'w')
            fd.write('test')
            fd.close()
            os.remove(fp)
        except IOError:
            return False
        else:
            return True

        if savesingle is not None and savesingle:
            savesingle = True
        else:
            settings = QtCore.QSettings()
            savesingle = settings.value('file/savesingle', True, bool)

        if not savesingle:
            fpd = self._related_dir(fp)
            if not os.path.exists(fpd):
                return False
            fp = os.path.join(fpd, 'test')
            return self.writeable(fp, False)
        else:
            return True

    def open(self, fp=None):
        '''read in a previously written set of files,
            or a single python source file.

        If the file is a lone python source file, no further processing
            will occur. However, if the file has been loaded in pynguin
            previously (it is being loaded from a .pyn or a _pynd):

        Any documents that look like function or class definitions will
            be exec'd to load them in to the interpreter local namespace.
        Any current history will be lost and replaced with the
            history loaded from the file.
        '''
        if not self.maybe_save():
            return

        if self._fdir is None:
            fdir = os.path.abspath(os.path.curdir)
        else:
            fdir = self._fdir

        logger.info('open at %s' % fdir)

        if not fp:
            fp, _ = QtWidgets.QFileDialog.getOpenFileName(self, 'Open file', fdir, 'Text files (*.pyn *.py)')

        if fp:
            if not os.path.exists(fp):
                QtWidgets.QMessageBox.information(self,
                            'Does not exist',
                            'File does not exist:\n\n%s' % fp)
                return

            fdir, _ = os.path.split(fp)

            oldfilepath = self._filepath
            if not fp.endswith('.py'):
                self._new()
                self._filepath = fp

            success = self._openfile(fp)
            if success:
                self._fdir = fdir
            else:
                self._filepath = oldfilepath

    def _update_after_open(self, fp, add_to_recent=False):
        self.ui.mselect.setCurrentIndex(0)
        self.changedoc(0)

        self._modified = False
        fdir, fname = os.path.split(fp)
        windowtitle = '%s [*] - Pynguin' % fname
        self.setWindowTitle(windowtitle)
        self.show_modified_status()

        if add_to_recent:
            self.addrecentfile(self._filepath)

    def _openfile(self, fp, add_to_recent=True, dump=False):
        '''Open the selected file.

        If the file is python source (.py) open the file and
            add it as a new page in the editor window.

        If the file is a pynguin archive (.pyn) open it and load
            its contents to the correct places.

            First determine which file format version is used in
            the file, then dispatch to the correct function.

        If dump=True, just prints out the various contents of the
            file to stdout if that is possible.

        '''
        if not os.path.exists(fp):
            QtWidgets.QMessageBox.information(self,
                    'File not found',
                    'File not found:\n\n%s' % fp)
            return False

        if fp.endswith('.pyn'):
            z = zipfile.ZipFile(fp, 'r')
            infos = z.infolist()
            info = infos[0]

            if not info.comment:
                ok = self._openfile00(fp, dump)
            else:
                if info.comment == b'pyn01':
                    ok = self._openfile01(fp, dump)
                else:
                    QtWidgets.QMessageBox.information(self,
                            'Unknown Format',
                            'Unknown file format:\n\n"%s"' % info.comment)
                    return False

        elif fp.endswith('.py'):
            if not dump:
                doc = self.editor.addexternal(fp)
                if doc is not None:
                    self.addrecentfile(fp)
                    self._modified = True
                    self.show_modified_status()
                    self._addwatcher(fp, doc)
            else:
                txt = open(fp).read()
                #txt = txt.decode('utf-8')
                print(txt)
                print()

            return True

        if ok:
            self._update_after_open(fp, add_to_recent)

        return True

    def _addwatcher(self, fp, doc):
        self.watcher.addPath(fp)
        self._watchdocs[str(fp)] = doc

    def _remwatcher(self, fp):
        self.watcher.removePath(fp)

    def _filechanged(self, fp):
        '''called when an external file has changed on disk.
        '''
        logger.info('_fc %s' % fp)

        settings = QtCore.QSettings()
        autorefresh = settings.value('file/reloadexternal', True, bool)
        if autorefresh:
            doc = self._watchdocs[str(fp)]
            if not os.path.exists(fp):
                self._modified = True
                self.show_modified_status()
                QtWidgets.QMessageBox.information(self,
                    'Moved',
                    'Local copy remains:\n\n%s\n\nSave to restore.' % fp)
                return

            if doc.isModified():
                QtWidgets.QMessageBox.information(self,
                    'Modified',
                    'Local copy has modifications:\n\n%s\n\nNot loading external changes.' % fp)
                return

            txt = open(fp).read()
            #txt = txt.decode('utf-8')
            doc.setText(txt)
            doc.setModified(False)
            autorun = settings.value('file/autorun', False, bool)
            if autorun:
                self.interpreter.runcode(txt)

        self.show_modified_status()

    def _remove_toplevel(self, data):
        '''When a page has both function definitions and top-level
            code, we only want to exec the function defs and not
            the top-level code when loading the file.

            This function takes the page of code and will return
            only the function and class defs with all other
            top-level code stripped out.
        '''

        newdata = []

        indef = False
        for line in data.splitlines():
            indented = line.startswith(' ')
            blank = line.isspace()
            if not line or blank:
                pass
            elif not indented:
                if line.startswith('def ') or line.startswith('class '):
                    indef = True
                else:
                    indef = False

            if indef:
                newdata.append(line)

        return '\n'.join(newdata)

    def _loaddata(self, data):
        if hasattr(data, 'decode'):
            data = data.decode('utf-8')
        self.editor.add(data)
        if data.startswith('def ') or data.startswith('class '):
            if data[4:12] == 'onclick(':
                # don't install onclick handlers when loading
                pass
            else:
                try:
                    data = self._remove_toplevel(data)
                    logger.info(data)
                    exec(data, self.interpreter_locals)
                except Exception as e:
                    print()
                    print('problem', e)
                    print('in...')
                    line1 = data.split('\n')[0]
                    print(str(line1))
                    self.interpretereditor.checkprompt()

    def _loadhistory(self, data):
        history = str(data).split('\n')
        self.interpretereditor.history = history

    def _shouldload00(self, ename):
        '''tries to load both original file format and any
            unknown format file.
        '''
        return ename.startswith('##')

    def _openfile00(self, fp, dump=False):
        'Used to open files from version < 0.10'
        z = zipfile.ZipFile(fp, 'r')
        for ename in z.namelist():
            fo = z.open(ename, 'r')
            data = fo.read()
            data = data.decode('utf-8')

            if dump:
                print(ename)
                print(data)
                print()
            elif self._shouldload00(ename):
                self._loaddata(data)
            elif ename.startswith('@@history@@'):
                self._loadhistory(data)

        return True

    def _shouldload01(self, fp):
        'return true if _openfile01 should load the file contents'
        while '\\' in fp:
            fp = fp.replace('\\', '/')
        while '//' in fp:
            fp = fp.replace('//', '/')
        edir, ename = os.path.split(fp)
        ok = len(ename)==8 and \
                ename[:5].isdigit() and \
                ename.endswith('.py')
        logger.info('SL %s %s' % (ok, fp))
        return ok

    def _openfile01(self, fp, dump=False):
        'used to open files in version 0.10 (file vers. pyn01)'
        fpd, _ = os.path.split(fp)
        z = zipfile.ZipFile(fp, 'r')
        namelist = z.namelist()
        logger.info(namelist)
        self._fformat = '.pyn'

        hasmanifest = False
        hashistory = False
        for ename in namelist:
            logger.info(ename)
            if '@@manifest@@' in ename:
                hasmanifest = ename
            elif '@@history@@' in ename:
                hashistory = ename

        if hasmanifest:
            fo = z.open(hasmanifest, 'r')
            data = fo.read()
            data = data.decode('utf-8')
            manifest_namelist = data.split('\n')
            if manifest_namelist:
                # Try opening the entries to see if the
                #   manifest is well formed or broken
                broken = False
                for ename in manifest_namelist:
                    if ename.startswith('_@@'):
                        # separate, internal (_pynd)
                        continue
                    if '_pynd' not in ename:
                        # manifest or history
                        continue

                    logger.info('Chk %s' % ename)

                    try:
                        fo = z.open(ename, 'r')
                    except KeyError:
                        #broken manifest
                        logger.info('Broken manifest %s %s' % (ename, manifest_namelist))
                        namelist.sort()
                        broken = True
                        break

                if not broken:
                    # Looks ok. Use the manifest
                    namelist = manifest_namelist
                    if hashistory:
                        namelist.append(hashistory)

        else:
            namelist.sort()

        if dump:
            print('Manifest/files:')
            for name in namelist:
                print('  ', name)

        pages = 0
        for ename in namelist:
            logger.info('loading %s' % ename)
            if dump:
                print('============================================')

            if ename.startswith('_@@'):
                # directory + .pyn style
                n = ename[4:]
                np = os.path.join(fpd, n)
                logger.info('NP %s' % np)
                npd, _ = os.path.split(np)
                if not os.path.exists(npd):
                    QtWidgets.QMessageBox.information(self,
                            'Not found',
'''File does not exist:

%s

This file requires an associated directory of files:

%s

Move this file and that folder to the same place and
try opening it again.

''' % (np, npd))
                    self.new()
                    return False
                try:
                    fo = open(np, 'r')
                except FileNotFoundError:
                    QtWidgets.QMessageBox.information(self,
                            'Not found',
'''File does not exist:

%s

Carrying on trying to recover if possible...

''' % np)
                else:
                    self._fformat = '_pynd'

            elif os.path.isabs(ename):
                try:
                    self._openfile(ename, dump=dump)
                    pages += 1
                    continue
                except IOError:
                    self.editor.add('Could not load:\n%s' % ename)
                    self.editor._doc._title = 'NOT LOADED'
                    self.editor._doc._filepath = ename
                    self.editor.settitle('NOT LOADED')
                    continue
            else:
                fo = z.open(ename, 'r')

            data = fo.read()
            logger.info('OPEN')
            logger.info(data)
            if hasattr(data, 'decode'):
                data = data.decode('utf-8')
            if dump:
                if '@@history@@' in ename:
                    print('Command history')
                print(data)
                print()
            elif self._shouldload01(ename):
                self._loaddata(data)
                pages += 1
            elif '@@history@@' in ename:
                self._loadhistory(data)

        if not pages:
            self.new()
        return True

    def _opendir(self, d):
        files = os.listdir(d)
        manifestname = '@@manifest@@'
        if manifestname in files:
            manifp = os.path.join(d, manifestname)
            maniff = open(manifp)
            manifest = maniff.read()
            files = manifest.split('\n')
            files = [os.path.join(d, f) for f in files if f]
        else:
            files.sort()

        for ename in files:
            fp = os.path.join(d, ename)
            fo = open(fp, 'r')
            data = fo.read()
            data = data.decode('utf-8')
            if self._shouldload01(ename):
                self._loaddata(data)
            elif ename == '@@history@@':
                self._loadhistory(data)

        self._update_after_open(d)

        return True

    def _verify_saved(self, fp):
        '''verify that the saved file contains the correct data.'''
        z = zipfile.ZipFile(fp, 'r')
        for ename in z.namelist():
            fo = z.open(ename, 'r')
            data = fo.read()
            data = data.decode('utf-8')
            if ename.startswith('##'):
                hdr = ename[0:9]
                if hdr.startswith('##') and hdr.endswith('##'):
                    docid = ename[11:]
                    if data != self.editor.documents[docid]:
                        QtWidgets.QMessageBox.warning(self,
                                'Unable to save',
                                'Files not saved!')
                        break

    def export(self, fp=None):
        '''save the current drawing.

        Format of saved image is determined by file suffix
            (default .png if none is given)

        '''

        if self._fdir is None:
            fdir = os.path.abspath(os.path.curdir)
        else:
            fdir = self._fdir

        if not fp:
            fp, _ = QtWidgets.QFileDialog.getSaveFileName(self, 'Export Image', fdir)
            fp = str(fp)
            if not fp:
                return False

        root, ext = os.path.splitext(fp)
        if not ext:
            ext = '.png'
            fp += ext
        fdir, _ = os.path.split(fp)

        for pynguin in self.pynguins:
            pynguin.gitem.hide()
            if hasattr(pynguin, '_pyn'):
                pynguin._pyn.gitem.hide()

        scene = self.scene
        view = scene.view

        ibr = scene.itemsBoundingRect()

        tl = ibr.topLeft()
        tlx, tly = tl.x(), tl.y()
        tlx -= 20
        tly -= 20
        tlp = QtCore.QPointF(tlx, tly)
        ibr.setTopLeft(tlp)

        br = ibr.bottomRight()
        brx, bry = br.x(), br.y()
        brx += 20
        bry += 20
        brp = QtCore.QPointF(brx, bry)
        ibr.setBottomRight(brp)

        scene.setSceneRect(scene.sceneRect().united(ibr))

        src = scene.sceneRect()

        if ext == '.svg':
            success = self.export_svg(fp, src)
        else:
            success = self.export_bitmap(fp, src)

        if success:
            self._fdir = fdir

        for pynguin in self.pynguins:
            pynguin.gitem.show()
            if hasattr(pynguin, '_pyn'):
                pynguin._pyn.gitem.show()

    def export_svg(self, fp, src):
        from PyQt5 import QtSvg

        svgen = QtSvg.QSvgGenerator()
        svgen.setFileName(fp)
        svgen.setSize(src.size().toSize())
        svgen.setTitle('Pynguin')
        painter = QtGui.QPainter(svgen)
        self.scene.render(painter)
        painter.end()

        return True

    def export_bitmap(self, fp, src):
        '''For .png .jpg or .tga'''

        sz = src.size().toSize()
        self._i = i = QtGui.QImage(sz, QtGui.QImage.Format_ARGB32)
        p = QtGui.QPainter(i)
        p.setRenderHint(QtGui.QPainter.Antialiasing, True)
        p.setRenderHint(QtGui.QPainter.TextAntialiasing, True)
        p.setRenderHint(QtGui.QPainter.SmoothPixmapTransform, True)
        ir = i.rect()
        irf = QtCore.QRectF(0, 0, src.width(), src.height())

        self.scene.render(p, irf, src)
        success = i.save(fp)#, quality=100)
        if not success:
            root, fn = os.path.split(fp)
            fnroot, ext = os.path.splitext(fn)
            if ext not in ('.png', '.jpg', '.jpeg', '.tga'):
                msg = 'Unsupported format.\n\nTry using %s.png' % fnroot
            else:
                msg = 'Cannot export image.'
            QtWidgets.QMessageBox.warning(self,
                                'Unable to save',
                                msg)

        return success

    def newdoc(self):
        '''Add a new (blank) page to the document editor'''
        docid = self.editor.new()
        self.editor.switchto(docid)
        self.editor._doc.setFocus()
        idx = self.ui.mselect.count() - 1
        idx = max(0, idx)
        self.ui.mselect.setCurrentIndex(idx)

        self._modified = True
        self.show_modified_status()

    def changedoc(self, idx):
        '''switch which document is visible in the document editor'''
        docid = str(self.ui.mselect.itemData(idx))
        if docid in self.editor.documents:
            self.editor.switchto(docid)
            self.editor._doc.setFocus()

    def removedoc(self):
        '''throw away the currently displayed editor document'''

        mselect = self.ui.mselect
        documents = self.editor.documents
        idx = mselect.currentIndex()
        docname = str(mselect.itemText(idx))
        docid = str(mselect.itemData(idx))
        doc = documents[docid]
        page_removed = self.editor.pages[docid]
        empty = not doc.text()
        logger.info('REMOVE')
        logger.info(str(documents))
        logger.info(idx)
        logger.info(docid)
        logger.info(self.editor.pages)
        logger.info(mselect.count())

        if hasattr(self.editor._doc, '_title'):
            external = True
            fp = self.editor._doc._filepath
        else:
            external = False

        modified = doc.isModified()

        if not modified and empty:
            affirm = True
        elif external and not modified:
            affirm = True
        else:
            ret = QtWidgets.QMessageBox.warning(self, 'Are you sure?',
                    'This page will be removed permanently.\n'
                    'Are you sure you want to remove this page?',
                    QtWidgets.QMessageBox.Yes|QtWidgets.QMessageBox.Cancel, QtWidgets.QMessageBox.Cancel)
            affirm = ret==QtWidgets.QMessageBox.Yes

        if not affirm:
            return
        else:
            mselect = self.ui.mselect
            idx = mselect.currentIndex()
            docname = str(mselect.itemText(idx))
            mselect.removeItem(idx)
            cwidg = self.editor.box.currentWidget()
            self.editor.box.removeWidget(cwidg)

            if docid in self.editor.documents:
                doc = self.editor.documents[docid]
                del self.editor.documents[docid]
                del self.editor.pages[docid]
                for pages_docid, pagen in self.editor.pages.items():
                    if pagen > page_removed:
                        self.editor.pages[pages_docid] = pagen - 1

            if mselect.count():
                self.changedoc(0)
                mselect.setCurrentIndex(0)
            else:
                self.editor._doc.setText('')
                self.newdoc()

            self._modified = True
            self.show_modified_status()

            if external:
                self._remwatcher(fp)

        logger.info('DONE')
        logger.info(self.editor.documents)
        logger.info(self.editor.pages)
        logger.info(mselect.count())

    def nextdoc(self):
        self.editor.shownext()

    def prevdoc(self):
        self.editor.showprev()

    def promotedoc(self):
        self.editor.promote()

    def demotedoc(self):
        self.editor.demote()

    def toggle_editor(self):
        if self.interpretereditor.hasFocus():
            self.editor.setFocus()
        else:
            self.interpretereditor.setFocus()

    def findmain(self, code):
        '''returns information about the main function or class in
            the given code.

        "Main" will depend on the settings and will either be the first
            function or class defined on the page, or the last one.

        returns a dictionary of:
            kind: 'function', 'class'
            name: function or class name
            params: string containing all parameters
            nodefault: list of params that do not have a default value
        '''

        settings = QtCore.QSettings()
        mainfirst = settings.value('editor/mainfirst', True, bool)
        lines = code.split('\n')
        if not mainfirst:
            lines.reverse()
        for line in lines:
            firstparen = line.find('(')
            lastparen = line.rfind(')')
            if line.startswith('def ') and line.endswith(':'):
                kind = 'function'
                name = line[4:firstparen]
            elif line.startswith('class ') and line.endswith(':'):
                kind = 'class'
                name = line[6:firstparen]
            else:
                continue

            if kind=='function' and firstparen > -1 and lastparen > -1:
                params = line[firstparen+1:lastparen]
                func = self.interpreter_locals.get(name, None)
                nodefault = []
                if func is not None:
                    defaults = util._plist(func)
                    if defaults:
                        for param, d in defaults:
                            if d is util.NODEFAULT:
                                nodefault.append(param)
            elif kind == 'class':
                c = self.interpreter_locals.get(name, None)
                nodefault = []
                params = ''
                if c is not None:
                    func = getattr(c, '__init__', None)
                    if func is not None:
                        defaults = util._plist(func)
                        if defaults:
                            del defaults[0] # self
                            count = len(defaults)
                            for i, (param, d) in enumerate(defaults):
                                params += param
                                if d is util.NODEFAULT:
                                    nodefault.append(param)
                                else:
                                    params += '=%s' % str(d)
                                if i < count-1:
                                    params += ', '

            return (kind, name, params, nodefault)

        return None, None, None, None

    def testcode(self):
        '''Activated by the Test/Run button (testrun)

            exec the code in the current editor window and load it in
            to the interpreter local namespace

            If the first line looks like a function definition, use
            it to feed a line to the interpreter set up to call the
            function.
        '''
        doc = self.editor._doc
        code = doc.cleancode()

        error = None

        try:
            compile(code, 'current file', 'exec')
        except SyntaxError as e:
            self.editor.selectline(e.lineno)
            self.interpretereditor.clearline()
            self.interpretereditor.write('Syntax Error in line %s\n' % e.lineno)
            self.interpretereditor.write('>>> ')
            self.editor.setFocus()
            error = 1
        else:
            self.interpretereditor.setFocus()
            if self.interpretereditor.cmdthread is None:
                self.interpretereditor.cmdthread = CmdThread(self.interpretereditor, code)
                cmdthread = self.interpretereditor.cmdthread
                cmdthread.start()
                self.watcherthread = WatcherThread(cmdthread)
                self.watcherthread.finished.connect(self.interpretereditor.testthreaddone)
                self.watcherthread.start()

                kind, name, params, nodefault = self.findmain(code)
                if kind is not None:
                    self.interpretereditor.clearline()
                    if kind == 'function':
                        tocall = '%s(%s)' % (name, params)
                    elif kind == 'class':
                        varname = name.lower()
                        if varname == name:
                            self.interpretereditor.write('# Class names should be capitalized \n')
                            self.interpretereditor.write('>>> ')
                            varname = varname[0]
                        tocall = '%s = %s(%s)' % (varname, name, params)

                    if kind=='function' and name == 'onclick':
                        self.interpretereditor.write('# set onclick handler\n')
                        self.interpretereditor.write('>>> ')
                    else:
                        self.interpretereditor.addcmd(tocall, force=True)
                else:
                    self.interpretereditor.write('\n')
            else:
                self.interpretereditor.write('not starting...\n')
                self.interpretereditor.write('code already running\n')
                #self.interpretereditor.write('>>> ')

        self.interpretereditor.setFocus()
        return error

    def testall(self):
        '''Test/Run each editor page in order.

        If any page throws an error, progress stops, the editor
        switches to that page and highlights the error.
        '''

        ie = self.interpretereditor
        if not ie.cmdthread is None:
            ie.write('not starting...2\n')
            ie.write('code already running\n')
            return

        settings = QtCore.QSettings()
        rev = settings.value('editor/testall_reverse', False, bool)
        autorun = settings.value('editor/testall_autocall', False, bool)
        reset = settings.value('editor/testrun_reset', True, bool)

        count = self.ui.mselect.count()
        Pynguin._stop_testall = False
        for i in range(count):
            if Pynguin._stop_testall:
                break

            if not rev:
                idx = i
            else:
                idx = count - i - 1
            docid = str(self.ui.mselect.itemData(idx))
            if docid in self.editor.documents:
                while ie.cmdthread is not None:
                    ie.spin(5)
                    if Pynguin._stop_testall:
                        break
                    Pynguin.wait_for_empty_q()
                    if self.pynguin.ControlC:
                        Pynguin._stop_testall = True
                        break
                self.editor.switchto(docid)
                self.editor.setFocus()
                if reset and not Pynguin._stop_testall:
                    ie.checkprompt()
                    ie.spin(5)
                    ie.addcmd('reset()')
                    ie.spin(5)
                    ie.go()

                while ie.cmdthread is not None:
                    ie.spin(5)
                    if Pynguin._stop_testall:
                        break
                    Pynguin.wait_for_empty_q()
                    if self.pynguin.ControlC:
                        Pynguin._stop_testall = True
                        break

                #if not ie.spin(0):
                    #Pynguin._stop_testall = True

                if Pynguin._stop_testall:
                    ie.clearline()
                    ie.write('Stopped\n')
                    ie.spin(5)
                    ie.checkprompt()
                    #ie.write('>>> ')
                    ie.setFocus()
                    break

                ie.clearline()
                ie.spin(5, delay=0.1)
                if self.testcode():
                    break
                ie.spin(5, delay=0.1)

                if autorun:
                    doc = self.editor.documents[docid]
                    code = doc.cleancode()
                    kind, name, params, nodefault = self.findmain(code)
                    missing_vars = []
                    if nodefault:
                        for var in nodefault:
                            if var not in self.interpreter_locals:
                                missing_vars.append(var)

                    if missing_vars:
                        ie.write('\n>>> ')
                        for var in missing_vars:
                            ie.write('# missing %s \n' % var)
                            ie.write('>>> ')
                        ie.spin(5)

                    elif not Pynguin._stop_testall:
                        ie.go()
                        logger.info('GO')
                        Pynguin.wait_for_empty_q()
                        logger.info('END WAIT')

        ie.clearline()

    def setPenColor(self):
        '''use a color selection dialog to set the pen line color

        sets the color for the primary pynguin only. For other later
            added pynguins, use p.color()
        '''
        icolor = self.pynguin.gitem.pen.brush().color()
        ncolor = QtWidgets.QColorDialog.getColor(
            icolor, self, 'Pen Color', QtWidgets.QColorDialog.ShowAlphaChannel)
        if ncolor.isValid():
            r, g, b, a = ncolor.getRgb()
            self.pynguin.color(r, g, b, a)
            if a != 255:
                cmd = 'color(%s, %s, %s, %s)\n' % (r, g, b, a)
            else:
                cmd = 'color(%s, %s, %s)\n' % (r, g, b)
            self.interpretereditor.addcmd(cmd)

    def set_background_color(self):
        '''Use color selection dialog to set the background color.
        '''

        settings = QtCore.QSettings()
        r, g, b = color = self.pynguin.bgcolor()
        c = QtGui.QColor(r, g, b)
        ncolor = QtWidgets.QColorDialog.getColor(c, self)
        if ncolor.isValid():
            r, g, b, _ = ncolor.getRgb()
            self.pynguin.bgcolor(r, g, b)
            cmd = 'bgcolor(%s, %s, %s)\n' % (r, g, b)
            self.interpretereditor.addcmd(cmd)

    def setPenWidth(self):
        '''open a dialog with a spin button to get a new pen line width

        sets the width for the primary pynguin only. For other later
            added pynguins, use p.width()
        '''
        iwidth = self.pynguin.gitem.pen.width()
        uifile = 'penwidth.ui'
        uipath = os.path.join(uidir, uifile)
        DClass, _ = uic.loadUiType(uipath)
        dc = DClass()
        d = QtWidgets.QDialog(self)
        d.ui = DClass()
        dc.setupUi(d)
        dc.thewid.setValue(iwidth)
        d.exec_()
        nwidth = dc.thewid.value()
        self.pynguin.width(nwidth)
        cmd = 'width(%s)\n' % nwidth
        self.interpretereditor.addcmd(cmd)

    def _setup_pendown_choices(self):
        choices = ((self.ui.actionPenUp, False),
                    (self.ui.actionPenDown, True))
        self._pendowns = bidict(choices)

    def _sync_pendown_menu(self, choice):
        action = self._pendowns.inv.get(choice)
        action.setChecked(True)

    def setPen(self, ev):
        '''toggle pen up and pen down

        sets the pen for the primary pynguin only. For other later
            added pynguins, use p.penup() or p.pendown()
        '''
        if ev == self.ui.actionPenUp:
            self.pynguin.penup()
            self.interpretereditor.addcmd('penup()\n')
        else:
            self.pynguin.pendown()
            self.interpretereditor.addcmd('pendown()\n')

    def _setup_fill_choices(self):
        choices = ((self.ui.actionFill, 'fill'),
                    (self.ui.actionNofill, 'nofill'))
        self._fills = bidict(choices)

    def _sync_fill_menu(self, choice):
        action = self._fills.inv.get(str(choice))
        action.setChecked(True)

    def setFill(self, ev):
        '''toggle fill on / off

        sets the fill for the primary pynguin only. For other later
            added pynguins, use p.fill() or p.nofill()
        '''
        if ev == self.ui.actionFill:
            self.pynguin.fill()
            self.interpretereditor.addcmd('fill()\n')
        else:
            self.pynguin.nofill()
            self.interpretereditor.addcmd('nofill()\n')

    def setFillColor(self):
        '''use a color selection dialog to set the fill color

        sets the fill color for the primary pynguin only. For other
            later added pynguins, use p.fillcolor()
        '''
        icolor = self.pynguin.gitem.brush.color()
        ncolor = QtWidgets.QColorDialog.getColor(icolor, self, 'Fill Color',
                                    QtWidgets.QColorDialog.ShowAlphaChannel)
        if ncolor.isValid():
            r, g, b, a = ncolor.getRgb()
            self.pynguin.fill()
            self._sync_fill_menu('fill')
            self.pynguin.fillcolor(r, g, b, a)
            if a != 255:
                cmd = 'fill(color=(%s, %s, %s, %s))\n' % (r, g, b, a)
            else:
                cmd = 'fill(color=(%s, %s, %s))\n' % (r, g, b)
            self.interpretereditor.addcmd(cmd)

    def setup_avatar_choices(self):
        choices = ((self.ui.actionPynguin, 'pynguin'),
                    (self.ui.actionArrow, 'arrow'),
                    (self.ui.actionRobot, 'robot'),
                    (self.ui.actionTurtle, 'turtle'),
                    (self.ui.actionHidden, 'hidden'))
        self.avatars = bidict(choices)

        custommenu = self.ui.menuCustom
        custommenu.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        # custommenu.connect(custommenu, QtCore.SIGNAL('customContextMenuRequested(const QPoint&)'), self.makedeleteaction())
        custommenu.customContextMenuRequested.connect(self.makedeleteaction())

    def _sync_avatar_menu(self, imageid, filepath=None):
        settings = QtCore.QSettings()
        if not filepath:
            # built-in avatar choices
            action = self.avatars.inv.get(str(imageid))
            if action is not None:
                action.setChecked(True)
            idpath = imageid
        else:
            fmt = QtGui.QImageReader.imageFormat(filepath)
            if fmt == 'svg':
                idpathstr = '@@_%s@@%s'
            else:
                idpathstr = '@@%s@@%s'
            idpath = idpathstr % (imageid, filepath)

            action = self.avatars.inv.get(idpath)
            if action is None:
                custommenu = self.ui.menuCustom
                action = custommenu.addAction(imageid)
                action.setCheckable(True)
                self.viewgroup.addAction(action)
                self.avatars[action:] = idpath
                self._save_custom_avatar(idpath)

            action.setChecked(True)

        settings.setValue('pynguin/avatar', idpath)

    def _save_custom_avatar(self, idpath, remove=False):
        settings = QtCore.QSettings()
        cavs = []
        n = settings.beginReadArray('pynguin/custom_avatars')
        for i in range(n):
            settings.setArrayIndex(i)
            cavs.append(settings.value('idpath'))
        settings.endArray()
        if not remove and idpath not in cavs:
            cavs.append(idpath)
        elif remove and idpath in cavs:
            cavs.remove(idpath)
        settings.beginWriteArray('pynguin/custom_avatars')
        for i, idp in enumerate(cavs):
            settings.setArrayIndex(i)
            settings.setValue('idpath', idp)
        settings.endArray()

        if remove:
            currid = settings.value('pynguin/avatar', 'pynguin')
            if currid == idpath:
                # Removed the current avatar
                self.set_pynguin_avatar('pynguin')

    def makedeleteaction(self):
        def deleteaction(pt, self=self):
            popup = QtWidgets.QMenu(self)
            custommenu = self.ui.menuCustom
            action = custommenu.actionAt(pt)
            def dodelete(action=action, menu=custommenu):
                menu.removeAction(action)
                idpath = self.avatars[action]
                self._save_custom_avatar(idpath, remove=True)
            popup.addAction('Remove', dodelete)
            popup.exec_(custommenu.mapToGlobal(pt))
        return deleteaction

    def set_pynguin_avatar(self, imageid, pyn=None):
        '''select which image to show

        sets the image for the primary pynguin only. For other later
            added pynguins, use p.avatar()
        '''

        if pyn is None:
            pyn = self.pynguin
            sync = None
        else:
            sync = False

        idpath = str(imageid)
        if idpath.startswith('@@_'):
            # custom svg
            imageid, filepath = idpath[3:].split('@@')
            imid = imageid
        elif idpath.startswith('@@'):
            # custom non-svg
            imageid, filepath = idpath[2:].split('@@')
            imid = None
        else:
            imid = imageid
            filepath = None
        pyn.avatar(imid, filepath)
        if sync:
            self._sync_avatar_menu(imageid, filepath)

        if not imid:
            imid = None
            cmdstr = "avatar(%s, '%s')\n"
            cmd = cmdstr % (imid, filepath)
        elif filepath is None:
            cmdstr = "avatar('%s')\n"
            cmd = cmdstr % imid
        else:
            cmdstr = "avatar('%s', '%s')\n"
            cmd = cmdstr % (imid, filepath)
        return cmd

    def setImageEvent(self, ev):
        imageid = self.avatars[ev]
        cmd = self.set_pynguin_avatar(imageid)
        self.interpretereditor.addcmd(cmd)

    def setcustomavatar(self):
        from . import avatar
        ad = avatar.CustomAvatar(self)
        r = ad.exec_()
        if r:
            filepath = str(ad.ui.filepath.text())
            element = str(ad.ui.element.text())
            if not element:
                element = None
                cmdstr = "avatar(%s, '%s')\n"
            else:
                cmdstr = "avatar('%s', '%s')\n"
            self.pynguin.avatar(element, filepath)

    def setup_speed_choices(self):
        choices = ((self.ui.actionSlow, 5),
                    (self.ui.actionMedium, 10),
                    (self.ui.actionFast, 20),
                    (self.ui.actionInstant, 0))
        self.speeds = bidict(choices)

    def sync_speed_menu(self, speed_value):
        action = self.speeds.inv.get(speed_value)
        if action is not None:
            action.setChecked(True)

    def speedMenuEvent(self, ev=None):
        if ev is None:
            ev = self.speedgroup.checkedAction()

        speed = self.speeds[ev]
        self.sync_speed_menu(speed)
        choices = { 5: 'slow',
                    10: 'medium',
                    20: 'fast',
                    0: 'instant'}

        if self.interpretereditor.cmdthread is not None:
            self.pynguin._speed(speed)

        choice = choices.get(speed)
        cmd = "speed('%s')\n" % choice
        self.interpretereditor.addcmd(cmd)
        self.pynguin.speed(choice)

    def wordwrap(self):
        checked = self.ui.actionWordwrap.isChecked()
        if not checked:
            self.editor.wrap(False)
        else:
            self.editor.wrap()
        settings = QtCore.QSettings()
        settings.setValue('editor/wordwrap', checked)

    def linenumbers(self):
        checked = self.ui.actionShowLineNumbers.isChecked()
        if checked:
            self.editor.line_numbers(True)
        else:
            self.editor.line_numbers(False)
        settings = QtCore.QSettings()
        settings.setValue('editor/linenumbers', checked)

    def zoomineditor(self):
        self.editor.zoomin()

    def zoomouteditor(self):
        self.editor.zoomout()

    def comment(self):
        self.editor._doc.commentlines()

    def uncomment(self):
        self.editor._doc.commentlines(un=True)

    def about(self):
        AboutDialog(self.app).exec_()

    def reportbug(self):
        QtGui.QDesktopServices().openUrl(QtCore.QUrl(conf.bug_url))

    def resizeEvent(self, ev):
        scene = self.scene
        size = ev.size()
        wnew, hnew = size.width(), size.height()
        rect = scene.sceneRect()
        wr, hr = rect.width(), rect.height()

        if wnew > wr or hnew > hr:
            w = max(wr, wnew)
            h = max(hr, hnew)
            w2, h2 = w/2, h/2
            scene.setSceneRect(-w2, -h2, w, h)

    def undo(self):
        self.pynguin._undo()


class Scene(QtWidgets.QGraphicsScene):
    def __init__(self):
        left = -300
        top = -300
        width = 600
        height = 600

        QtWidgets.QGraphicsScene.__init__(self)
        self.setSceneRect(left, top, width, height)

        #self.bg = QtGui.QPixmap(width+1, height+1)
        #self.bg.fill(bgcolor)
        #self.bgp = QtGui.QPainter(self.bg)
        #self.bgp.drawEllipse(QtCore.QRect(0, 0, width, height))

    def drawBackground2(self, painter, rect):
        '''can be used in conjunction with background pixmap and
            painter set up in __init__ to draw an image behind
            the scene items.

            Not currently used.
        '''
        QtWidgets.QGraphicsScene.drawBackground(self, painter, rect)
        rt = QtCore.QRectF(rect)
        rtl = rect.topLeft()
        x, y = rtl.x(), rtl.y()
        pt = QtCore.QPointF(x+300, y+300)
        rt.moveTopLeft(pt)
        painter.drawPixmap(rect, self.bg, rt)
