# Copyright 2010-2018 Lee Harr
#
# This file is part of pynguin.
# https://bitbucket.org/leeharr/pynguin/
#
# Pynguin is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pynguin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pynguin.  If not, see <http://www.gnu.org/licenses/>.


import os
import sys
from random import randrange, choice
from math import copysign

from PyQt5 import QtCore, QtSvg, QtGui
from PyQt5.QtGui import QColor

import logging
logger = logging.getLogger('PynguinLogger')


NOTHING = object()

def sign(x):
    'return 1 if x is positive, -1 if negative, or zero'
    return copysign(1, x)


NODEFAULT = object()
def _plist(f):
    '''given function object,
        return a list of [(arg name: default value or None), ...]
    '''
    parameter_defaults = []
    defaults = f.__defaults__
    if defaults is not None:
        defaultcount = len(defaults)
    else:
        defaultcount = 0
    argcount = f.__code__.co_argcount
    for i in range(f.__code__.co_argcount):
        name = f.__code__.co_varnames[i]
        value = NODEFAULT
        if i >= argcount - defaultcount:
            value = defaults[i - (argcount - defaultcount)]
        parameter_defaults.append((name, value))
    return parameter_defaults


class SvgRenderer(object):
    'factory for svg renderer objects'

    def __init__(self, app):
        self.app = app

    def getrend(self, filepath=None):
        '''return a handle to the shared SVG renderer
            for the given svg file.

        If no filepath is given, return the renderer for
            the default svg file.

        '''
        if filepath is None:
            filename = 'pynguin.svg'
            filepath = os.path.join(datadir, 'images', filename)
        fp = filepath
        rend = QtSvg.QSvgRenderer(fp, self.app)
        return rend


def choose_color(*args, **kw):
    '''choose_color version 2, usable many ways:

    choose_color('<color name>')
    choose_color('random' | 'rlight' | 'rmedium' | 'rdark')
    choose_color(random=True | 'light' | 'medium' | 'dark')
    choose_color(name='<color name>')

    choose_color(r, g, b)
    choose_color(r, g, b, a)
    choose_color(r=<val>, g=<val>, b=<val>) # val will be clamped to int(0)-int(255)
    choose_color(None) # same as cc(a=0)

    choose_color(<3-tuple>) # (r, g, b)
    choose_color(<4-tuple>) # (r, g, b, a)

    choose_color('<Qt Color>')

    >>> color_names = ('red', 'green', 'blue', 'orange')
    >>> for name in color_names:
    ...     choose_color(name)
    (255, 0, 0, 255)
    (0, 128, 0, 255)
    (0, 0, 255, 255)
    (255, 165, 0, 255)

    >>> choose_color('notacolor')
    Traceback (most recent call last):
        ...
    ValueError


    >>> import PyQt5
    >>> import PyQt5.QtGui
    >>> from PyQt5.QtGui import QColor
    >>> for name in color_names:
    ...     qc = QColor(name)
    ...     choose_color(qc)
    (255, 0, 0, 255)
    (0, 128, 0, 255)
    (0, 0, 255, 255)
    (255, 165, 0, 255)


    >>> for name in color_names:
    ...     choose_color(name=name)
    (255, 0, 0, 255)
    (0, 128, 0, 255)
    (0, 0, 255, 255)
    (255, 165, 0, 255)


    >>> import random
    >>> random.seed(0)
    >>> for rname in ('random', 'rlight', 'rmedium', 'rdark'):
    ...     choose_color(rname)
    (197, 215, 20, 255)
    (216, 232, 231, 255)
    (151, 138, 161, 255)
    (45, 74, 27, 255)

    >>> random.seed(0)
    >>> for rname in (True, 'light', 'medium', 'dark'):
    ...     choose_color(random=rname)
    (197, 215, 20, 255)
    (216, 232, 231, 255)
    (151, 138, 161, 255)
    (45, 74, 27, 255)

    >>> color_tups = [(1, 2, 3), (2, 3, 4, 5), (-1, 300, 255)]
    >>> for tup in color_tups:
    ...     choose_color(*tup)
    (1, 2, 3, 255)
    (2, 3, 4, 5)
    (0, 255, 255, 255)

    >>> choose_color(1, 2)
    Traceback (most recent call last):
        ...
    ValueError
    >>> choose_color(1, 2, 3, 4, 5)
    Traceback (most recent call last):
        ...
    ValueError


    >>> for tup in color_tups:
    ...     choose_color(tup)
    (1, 2, 3, 255)
    (2, 3, 4, 5)
    (0, 255, 255, 255)

    >>> choose_color((1, 2))
    Traceback (most recent call last):
        ...
    ValueError
    >>> choose_color((1, 2, 3, 4, 5))
    Traceback (most recent call last):
        ...
    ValueError


    >>> choose_color(r=10)
    (10, None, None, None)
    >>> choose_color(g=20)
    (None, 20, None, None)
    >>> choose_color(b=30)
    (None, None, 30, None)
    >>> choose_color(a=40)
    (None, None, None, 40)
    >>> choose_color(r=340)
    (255, None, None, None)

    >>> choose_color(r=5, g=15)
    (5, 15, None, None)
    >>> choose_color(r=5, b=25)
    (5, None, 25, None)
    >>> choose_color(r=5, g=15, b=25)
    (5, 15, 25, None)
    >>> choose_color(r=5, g=15, b=25, a=35)
    (5, 15, 25, 35)


    >>> choose_color(None)
    (None, None, None, 0)

    '''

    name = None
    rand = None
    tup3 = None
    tup4 = None
    qcol = None

    randranges = {
        'random': range(256),
        'rlight': range(200, 256),
        'rmedium': range(100, 200),
        'rdark': range(100),
        'ralpha': range(256)}

    if args and kw:
        raise ValueError
    elif len(args)==1 or 'random' in kw:
        if args and len(args)==1:
            arg0 = args[0]
        else:
            arg0 = None

        if type(arg0)==QColor:
            qcol = arg0
        elif args and arg0 is None:
            kw = dict(r=None, g=None, b=None, a=0)
        elif arg0 in randranges:
            rand = arg0
        elif type(arg0)==tuple and len(arg0)==3:
            tup3 = arg0
        elif type(arg0)==tuple and len(arg0)==4:
            tup4 = arg0
        elif type(arg0)==tuple and (len(arg0)==2 or len(arg0)>4):
            raise ValueError
        elif 'random' in kw:
            kwr = kw['random']
            if kwr is True:
                kwr = 'andom'
            elif not kwr:
                raise ValueError
            rand = 'r' + kwr
        elif QColor.isValidColor(arg0):
            name = arg0
    elif 'name' in kw:
        kwn = kw['name']
        if kwn and QColor.isValidColor(kwn):
            name = kwn
        else:
            raise ValueError
    elif len(args)==3:
        tup3 = args
    elif len(args)==4:
        tup4 = args


    if rand is not None:
        rg = randranges[rand]
        r, g, b = (choice(rg) for i in range(3))
        if rand == 'ralpha':
            a = choice(range(100, 200))
        else:
            a = 255

    elif name is not None:
        c = QtGui.QColor(name)
        r, g, b, a = c.red(), c.green(), c.blue(), c.alpha()

    elif qcol is not None:
        c = QtGui.QColor(qcol)
        r, g, b, a = c.red(), c.green(), c.blue(), c.alpha()

    elif kw:
        r = kw.get('r', None)
        g = kw.get('g', None)
        b = kw.get('b', None)
        a = kw.get('a', None)

    elif tup3 is not None:
        r, g, b = tup3
        a = 255

    elif tup4 is not None:
        r, g, b, a = tup4

    else:
        raise ValueError

    if r is not None:
        r = max(0, min(255, r))
    if g is not None:
        g = max(0, min(255, g))
    if b is not None:
        b = max(0, min(255, b))
    if a is not None:
        a = max(0, min(255, a))

    return r, g, b, a

def nudge_color(color, r=None, g=None, b=None, a=None):
    """Change the color (a 3-element or 4-element tuple) by given amounts,
        return the new RGBA tuple.

    Clamps the RGBA return values such that 0 <= RGBA <= 255
        but does not necessarily return only integer values.

        Not returning strictly integers allows for smoother color
        variations, but note that when the values are passed
        to the pynguin color() function the values will be
        converted to integers. So in order to take advantage
        of the more precise values you will need to keep those
        separately from the actual pynguin color values.

    The function's r, g, b, a parameters can be either:

        numbers to be added to or subtracted from the RGB tuple
            components, or

        percentages (as strings) that will be multiplied by the component
            to increase or decrease that component by given the given
            percent.

    >>> color = (100, 100, 100)
    >>> nudge_color(color, g=15)
    (100, 115, 100, 255)

    >>> color = (100, 100, 100)
    >>> nudge_color(color, r=-12.5)
    (87.5, 100, 100, 255)

    >>> color = (100, 100, 100)
    >>> color = nudge_color(color, b='75%')
    >>> color
    (100, 100, 75.0, 255)
    >>> nudge_color(color, b='75%')
    (100, 100, 56.25, 255)

    >>> color = (100, 100, 100)
    >>> nudge_color(color, r=50, g='105%', b=-10)
    (150, 105.0, 90, 255)
    """

    if len(color) == 3:
        rc, gc, bc = color
        ac = 255
    elif len(color) == 4:
        rc, gc, bc, ac = color
    else:
        raise ValueError

    if r is not None:
        try:
            rc += r
        except TypeError as e:
            p = r[-1:]
            if p != '%':
                raise ValueError("String value must be a percent (ending in '%').") from e
            rc *= (float(r[:-1]) / 100.0)

    if g is not None:
        try:
            gc += g
        except TypeError as e:
            p = g[-1:]
            if p != '%':
                raise ValueError("String value must be a percent (ending in '%').") from e
            gc *= (float(g[:-1]) / 100.0)

    if b is not None:
        try:
            bc += b
        except TypeError as e:
            p = b[-1:]
            if p != '%':
                raise ValueError("String value must be a percent (ending in '%').") from e
            bc *= (float(b[:-1]) / 100.0)

    if a is not None:
        try:
            ac += a
        except TypeError:
            ac *= (float(a[:-1]) / 100.0)

    rc = min(rc, 255)
    gc = min(gc, 255)
    bc = min(bc, 255)
    ac = min(ac, 255)
    rc = max(rc, 0)
    gc = max(gc, 0)
    bc = max(bc, 0)
    ac = max(ac, 0)

    return (rc, gc, bc, ac)


def get_datadir():
    if '-d' in sys.argv:
        dd = 'data'
        base = ''
    else:
        dd = 'share/pynguin'
        base = sys.prefix

    absbase = os.path.abspath(base)

    return os.path.join(absbase, dd)

def get_docdir():
    if '-d' in sys.argv:
        dd = 'doc'
        base = ''
    else:
        dd = 'share/doc/pynguin'
        base = sys.prefix

    absbase = os.path.abspath(base)

    return os.path.join(absbase, dd)

def check_data_doc_dir():
    d1 = get_datadir()
    d2 = get_docdir()
    return os.path.exists(d1) and os.path.exists(d2)

datadir = get_datadir()


if __name__ == '__main__':
    import doctest
    doctest.testmod()
